{% extends 'base.html' %}

{% block announce %}
<span class="twemoji" style="vertical-align: middle;">
  {% include ".icons/octicons/heart-fill-24.svg" %}
</span>
<a class="pdm-expansions" style="display: none;">Python Dependency Manager</a>
{% endblock %}
